@@ -14,7 +14,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <GL/glew.h>
-#include <GL/glut.h>
+#include "glut_wrap.h"
 
 /* For debug */
 #define DEPTH 1
