@@ -37,7 +37,7 @@
 #include <string.h>
 #include <math.h>
 #include <GL/glew.h>
-#include <GL/glut.h>
+#include "glut_wrap.h"
 
 #define DEG_TO_RAD (3.14159 / 180.0)
 
