@@ -525,7 +525,7 @@ event_loop(Display *dpy, GLXWindow win)
          case KeyPress:
             {
                char buffer[10];
-               int r, code;
+               int code;
                code = XLookupKeysym(&event.xkey, 0);
                if (code == XK_Left) {
                   view_roty += 5.0;
@@ -540,8 +540,8 @@ event_loop(Display *dpy, GLXWindow win)
                   view_rotx -= 5.0;
                }
                else {
-                  r = XLookupString(&event.xkey, buffer, sizeof(buffer),
-                                    NULL, NULL);
+                  XLookupString(&event.xkey, buffer, sizeof(buffer),
+                                NULL, NULL);
                   if (buffer[0] == 27) {
                      /* escape */
                      return;
@@ -573,6 +573,7 @@ event_loop(Display *dpy, GLXWindow win)
             GLfloat fps = frames / seconds;
             printf("%d frames in %3.1f seconds = %6.3f FPS\n", frames, seconds,
                    fps);
+            fflush(stdout);
             t0 = t;
             frames = 0;
          }
